@@ -28,7 +28,7 @@ import qualified Data.Text as T
 import qualified Graphics.Vty as V
 import           Lens.Micro.Platform ( (.~), to )
 import           Network.Mattermost.Lenses ( postEditAtL, postCreateAtL )
-import           Network.Mattermost.Types ( ServerTime(..), userUsername )
+import           Network.Mattermost.Types ( ServerTime(..), userUsername, postId )
 import           Prelude ()
 import           Matterhorn.Prelude
 
@@ -125,8 +125,12 @@ renderChatMessage st hs ind threadState renderTimeFunc msg =
 -- because (depending on the situation) we might use either of the
 -- message list types for the 'before' and 'after' (i.e. the
 -- chronological or retrograde message sequences).
-unsafeRenderMessageSelection :: (Foldable f, Foldable g)
-                             => ((Message, ThreadState), (f (Message, ThreadState), g (Message, ThreadState)))
+unsafeRenderMessageSelection :: (SeqDirection dir1, SeqDirection dir2)
+                             => ( (Message, ThreadState)
+                                , ( DirectionalSeq dir1 (Message, ThreadState)
+                                  , DirectionalSeq dir2 (Message, ThreadState)
+                                  )
+                                )
                              -> (Message -> ThreadState -> Widget Name)
                              -> Widget Name
 unsafeRenderMessageSelection ((curMsg, curThreadState), (before, after)) doMsgRender =
@@ -140,24 +144,54 @@ unsafeRenderMessageSelection ((curMsg, curThreadState), (before, after)) doMsgRe
         upperHeight = targetHeight `div` 2
         lowerHeight = targetHeight - upperHeight
 
-        lowerRender img (m, tState) = render1HLimit doMsgRender V.vertJoin targetHeight img tState m
-        upperRender img (m, tState) = render1HLimit doMsgRender (flip V.vertJoin) targetHeight img tState m
+    lowerHalfResults <- renderMessageSeq targetHeight (render1 doMsgRender) vLimit after
+    upperHalfResults <- renderMessageSeq targetHeight (render1 doMsgRender) cropTopTo before
 
-    lowerHalf <- foldM lowerRender V.emptyImage after
-    upperHalf <- foldM upperRender V.emptyImage before
+    let upperHalfResultsHeight = sum $ (V.imageHeight . image) <$> upperHalfResults
+        lowerHalfResultsHeight = sum $ (V.imageHeight . image) <$> lowerHalfResults
+        curHeight = V.imageHeight $ curMsgResult^.imageL
+        uncropped = vBox $ fmap resultToWidget $
+                           (reverse upperHalfResults) <> (curMsgResult : lowerHalfResults)
 
-    let curHeight = V.imageHeight $ curMsgResult^.imageL
-        uncropped = upperHalf V.<-> curMsgResult^.imageL V.<-> lowerHalf
-        img = if | V.imageHeight lowerHalf < (lowerHeight - curHeight) ->
-                     V.cropTop targetHeight uncropped
-                 | V.imageHeight upperHalf < upperHeight ->
-                     V.cropBottom targetHeight uncropped
-                 | otherwise ->
-                     V.cropTop upperHeight upperHalf V.<-> curMsgResult^.imageL V.<->
-                        (if curHeight < lowerHeight
-                          then V.cropBottom (lowerHeight - curHeight) lowerHalf
-                          else V.cropBottom lowerHeight lowerHalf)
-    return $ emptyResult & imageL .~ img
+        cropTop h w = Widget Fixed Fixed $ do
+            result <- withReaderT relaxHeight $ render w
+            render $ cropTopTo h $ resultToWidget result
+        cropBottom h w = Widget Fixed Fixed $ do
+            result <- withReaderT relaxHeight $ render w
+            render $ cropBottomTo h $ resultToWidget result
+
+        lowerHalf = vBox $ fmap resultToWidget lowerHalfResults
+        upperHalf = vBox $ fmap resultToWidget $ reverse upperHalfResults
+
+    render $ if | lowerHalfResultsHeight < (lowerHeight - curHeight) ->
+                    cropTop targetHeight uncropped
+                | upperHalfResultsHeight < upperHeight ->
+                    vLimit targetHeight uncropped
+                | otherwise ->
+                    cropTop upperHeight upperHalf <=> (resultToWidget curMsgResult) <=>
+                       (if curHeight < lowerHeight
+                         then cropBottom (lowerHeight - curHeight) lowerHalf
+                         else cropBottom lowerHeight lowerHalf)
+
+resultToWidget :: Result n -> Widget n
+resultToWidget = Widget Fixed Fixed . return
+
+renderMessageSeq :: (SeqDirection dir)
+                 => Int
+                 -> (Message -> ThreadState -> Widget Name)
+                 -> (Int -> Widget Name -> Widget Name)
+                 -> DirectionalSeq dir (Message, ThreadState)
+                 -> RenderM Name [Result Name]
+renderMessageSeq remainingHeight renderFunc limitFunc ms
+    | messagesLength ms == 0 = return []
+    | otherwise = do
+        let Just (m, threadState) = messagesHead ms
+            maybeCache = case m^.mMessageId of
+                Nothing -> id
+                Just i -> cached (RenderedMessage i)
+        result <- render $ limitFunc remainingHeight $ maybeCache $ renderFunc m threadState
+        rest <- renderMessageSeq (remainingHeight - (V.imageHeight $ result^.imageL)) renderFunc limitFunc (messagesDrop 1 ms)
+        return $ result : rest
 
 renderLastMessages :: ChatState
                    -> HighlightSet
@@ -175,54 +209,55 @@ renderLastMessages st hs editCutoff msgs =
 
             isBelow m transition = m^.mDate > transition^.mDate
 
-            go :: V.Image -> DirectionalSeq Retrograde (Message, ThreadState) -> RenderM Name V.Image
-            go img ms | messagesLength ms == 0 = return img
-            go img ms = do
+            go :: Int -> DirectionalSeq Retrograde (Message, ThreadState) -> RenderM Name [Result Name]
+            go _ ms | messagesLength ms == 0 = return []
+            go remainingHeight ms = do
                 let Just (m, threadState) = messagesHead ms
                     newMessagesAbove = maybe False (isBelow m) newMessageTransition
-                newImg <- render1HLimit doMsgRender (flip V.vertJoin) targetHeight img threadState m
+
+                result <- render $ render1 doMsgRender m threadState
+
+                croppedResult <- render $ cropTopTo remainingHeight $ resultToWidget result
+
                 -- If the new message fills the window, check whether
                 -- there is still a "New Messages" transition that is
                 -- not displayed. If there is, then we need to replace
                 -- the top line of the new image with a "New Messages"
                 -- indicator.
-                if V.imageHeight newImg >= targetHeight && newMessagesAbove
+                if V.imageHeight (result^.imageL) >= remainingHeight
                 then do
-                    transitionResult <- render $ withDefAttr newMessageTransitionAttr $
-                                                 hBorderWithLabel (txt "New Messages ↑")
-                    let newImg2 = V.vertJoin (transitionResult^.imageL)
-                                               (V.cropTop (targetHeight - 1) newImg)
-                    return newImg2
-                else go newImg $ messagesDrop 1 ms
+                    single <- if newMessagesAbove
+                              then do
+                                  result' <- render $
+                                      vBox [ withDefAttr newMessageTransitionAttr $ hBorderWithLabel (txt "New Messages ↑")
+                                           , cropTopBy 1 $ resultToWidget croppedResult
+                                           ]
+                                  return result'
+                              else do
+                                  return croppedResult
+                    return [single]
+                else do
+                    let unusedHeight = remainingHeight - V.imageHeight (result^.imageL)
+                    rest <- go unusedHeight $ messagesDrop 1 ms
+                    return $ result : rest
 
-        img <- go V.emptyImage msgs
-        return $ emptyResult & imageL .~ (V.cropTop targetHeight img)
+        results <- go targetHeight msgs
+        render $ vBox $ (Widget Fixed Fixed . return) <$> reverse results
 
 relaxHeight :: Context -> Context
 relaxHeight c = c & availHeightL .~ (max maxMessageHeight (c^.availHeightL))
 
-render1HLimit :: (Message -> ThreadState -> Widget Name)
-              -> (V.Image -> V.Image -> V.Image)
-              -> Int
-              -> V.Image
-              -> ThreadState
-              -> Message
-              -> RenderM Name V.Image
-render1HLimit doMsgRender fjoin lim img threadState msg
-  | V.imageHeight img >= lim = return img
-  | otherwise = fjoin img <$> render1 doMsgRender threadState msg
-
 render1 :: (Message -> ThreadState -> Widget Name)
-        -> ThreadState
         -> Message
-        -> RenderM Name V.Image
-render1 doMsgRender threadState msg = case msg^.mDeleted of
-    True -> return V.emptyImage
-    False -> do
-        r <- withReaderT relaxHeight $
-             render $ padRight Max $
-             doMsgRender msg threadState
-        return $ r^.imageL
+        -> ThreadState
+        -> Widget Name
+render1 doMsgRender msg threadState = case msg^.mDeleted of
+    True -> emptyWidget
+    False ->
+        Widget Greedy Fixed $ do
+            withReaderT relaxHeight $
+                render $ padRight Max $
+                doMsgRender msg threadState
 
 -- | A bundled structure that includes all the information necessary
 -- to render a given message
@@ -281,18 +316,24 @@ renderMessage md@MessageData { mdMessage = msg, .. } =
           Just u -> if omittedUsernameType (msg^.mType) then Nothing else Just u
           Nothing -> Nothing
         botElem = if isBotMessage msg then txt "[BOT]" else emptyWidget
+        mId = msg^.mMessageId
+        clickableAuthor un = case mId of
+            Nothing -> id
+            -- We use the index (-1) since indexes for clickable
+            -- usernames elsewhere in this message start at 0.
+            Just i -> clickable (ClickableUsernameInMessage i (-1) un)
         nameElems = case msgUsr of
           Just un
             | isEmote msg ->
                 [ withDefAttr pinnedMessageIndicatorAttr $ txt $ if msg^.mPinned then "[PIN]" else ""
                 , txt $ (if msg^.mFlagged then "[!] " else "") <> "*"
-                , colorUsername mdMyUsername un un
+                , clickableAuthor un $ colorUsername mdMyUsername un un
                 , botElem
                 , txt " "
                 ]
             | otherwise ->
                 [ withDefAttr pinnedMessageIndicatorAttr $ txt $ if msg^.mPinned then "[PIN] " else ""
-                , colorUsername mdMyUsername un un
+                , clickableAuthor un $ colorUsername mdMyUsername un un
                 , botElem
                 , txt $ (if msg^.mFlagged then "[!]" else "") <> ": "
                 ]
@@ -331,13 +372,36 @@ renderMessage md@MessageData { mdMessage = msg, .. } =
                  ]
         msgReac = if Map.null (msg^.mReactions) || (not mdShowReactions)
           then Nothing
-          else let renderR e us =
+          else let renderR e us lst =
                        let n = Set.size us
-                       in if | n == 1    -> " [" <> e <> "]"
-                             | n > 1     -> " [" <> e <> " " <> T.pack (show n) <> "]"
-                             | otherwise -> ""
-                   reactionMsg = Map.foldMapWithKey renderR (msg^.mReactions)
-               in Just $ withDefAttr emojiAttr $ txt ("   " <> reactionMsg)
+                       in if | n == 1    -> makeReactionWidget e us (" [" <> e <> "]") : lst
+                             | otherwise -> makeReactionWidget e us (" [" <> e <> " " <> T.pack (show n) <> "]") : lst
+                   nonEmptyReactions = Map.filter (not . Set.null) $ msg^.mReactions
+                   makeReactionWidget e us t =
+                       let w = txt t in
+                       maybe w (flip clickable w) $ makeName e us
+                   hasAnyReactions = not $ null nonEmptyReactions
+                   makeName e us = do
+                       pid <- postId <$> msg^.mOriginalPost
+                       Just $ ClickableReactionInMessage pid e us
+                   reactionWidget = Widget Fixed Fixed $ do
+                       ctx <- getContext
+                       let lineW = ctx^.availWidthL
+                       reacs <- mapM render $ Map.foldrWithKey renderR [] nonEmptyReactions
+                       let reacLines :: [Result n] -> Int -> [Result n] -> [[Result n]]
+                           reacLines l _ []     = if null l then [] else [l]
+                           reacLines l w (r:rs) =
+                               let rW = V.imageWidth $ r^.imageL
+                               in if rW <= w
+                                  then reacLines (l <> [r]) (w - rW) rs
+                                  else
+                                      let rest = reacLines [] lineW rs
+                                      in l : [r] : rest
+
+                       render $ vBox $ hBox <$> (fmap (fmap resultToWidget)) (reacLines [] lineW reacs)
+               in if hasAnyReactions
+                  then Just $ withDefAttr emojiAttr $ txt "   " <+> reactionWidget
+                  else Nothing
         withParent p =
             case mdThreadState of
                 NoThread -> msgWidget
@@ -377,7 +441,7 @@ renderMessage md@MessageData { mdMessage = msg, .. } =
             if mdIndentBlocks
                then vBox [ hBox nameElems
                          , hBox [txt "  ", renderRichText mdMyUsername hs ((subtract 2) <$> w)
-                                                 mdWrapNonhighlightedCodeBlocks (Blocks bs)]
+                                                 mdWrapNonhighlightedCodeBlocks (Just clickableNames) (Blocks bs)]
                          ]
                else nameNextToMessage hs w nameElems bs
 
@@ -385,12 +449,22 @@ renderMessage md@MessageData { mdMessage = msg, .. } =
             Widget Fixed Fixed $ do
                 nameResult <- render $ hBox nameElems
                 let newW = subtract (V.imageWidth (nameResult^.imageL)) <$> w
-                render $ hBox [ raw (nameResult^.imageL)
-                              , renderRichText mdMyUsername hs newW mdWrapNonhighlightedCodeBlocks (Blocks bs)
+                render $ hBox [ Widget Fixed Fixed $ return nameResult
+                              , renderRichText mdMyUsername hs newW mdWrapNonhighlightedCodeBlocks (Just clickableNames) (Blocks bs)
                               ]
 
         breakCheck i = i `elem` [ELineBreak, ESoftBreak]
 
+        clickableNames i (EHyperlink u _) =
+            case msg^.mMessageId of
+                Just mId -> Just $ ClickableURLInMessage mId i $ LinkURL u
+                Nothing -> Nothing
+        clickableNames i (EUser name) =
+            case msg^.mMessageId of
+                Just mId -> Just $ ClickableUsernameInMessage mId i name
+                Nothing -> Nothing
+        clickableNames _ _ = Nothing
+
 -- Add the edit sentinel to the end of the last block in the sequence.
 -- If the last block is a paragraph, append it to that paragraph.
 -- Otherwise, append a new block so it appears beneath the last
